# Bing search functions
# (c) 12/2023, J. Mazanec

# extract abs nr of occurrences  
 extract.nr <- function(nr){
 nr <- parse_number(nr)
 nr <- as.integer(nr)
 return(nr)
} 

# capture frequencies for a keyword vector
gsearch.vec <- function(rowfn, context="", outfn="", repet = 1){
 # browser()
 rowkw <- scan(rowfn, what = 'character')
 rlgth <- length(rowkw)
 Frequ <- matrix(c(rep(0,rlgth*repet)), nrow = rlgth)
 rownames(Frequ) <- rowkw
 colnames(Frequ) <- "abs frequ"
 urlname1 <- 'https://www.bing.com/search?q='
 # qu <- ""     #"%22"=quotation mark  %20=space
 r <- 1
 while(r < repet + 1){
  i <- 1
  while(i < rlgth + 1){
    closeAllConnections()
    tmprkw <- gsub("[[:space:]]", "+", rowkw[i])	
    url <- paste(urlname1,tmprkw,"+",context,sep="")
    # print(url)
    # url.show(url) 
    s <- read_html(url)
    p <- s %>% html_elements("body")
    # Sys.sleep(5)
    p <- as(p, "character")
    writeLines(p, 'tmp')
  
    q <- scan('tmp','character')
    loc <- which(q == 'results</span></div>')
    if (identical(loc, integer(0))) { loc <- which(q == 'Ergebnisse</span></div>') }
    nr <- paste(str_trim(q[loc-3]), str_trim(q[loc-2]), str_trim(q[loc-1]), sep="")
    if (str_length(q[loc-4]) <= 3) { nr <- paste(str_trim(q[loc-4]), nr, sep="") }

    nr <- extract.nr(nr)
    # print(nr)
    # if (length(loc) == 0) {nr <- 0}
    Frequ[i,r] <- Frequ[i,r] + nr 			
    unlink('tmp')
    unlink(url)
    i <- i + 1
  }
  r <- r + 1
  }
 closeAllConnections()
 if (outfn != ""){
  Frequ[is.na(Frequ)] <- 1				# replace NAs by 1
  write(t(Frequ), outfn, ncolumns = repet, sep=" ")
  cat("\nVector written to file ",outfn,";\n",sep="")
  cat("zero occurrences set to unity:\n")
 } 
 print(Frequ)
 return(Frequ)
}

# capture frequencies for a keyword matrix
gsearch.mat <- function(rowfn, colfn, context="", outfn=""){
 sym = FALSE
 if (rowfn == colfn) {sym = TRUE}
 rowkw <- scan(rowfn, what = 'character')
 rlgth <- length(rowkw)
 colkw <- scan(colfn, what = 'character')
 clgth <- length(colkw)
 cat("\nLaunching",rlgth*clgth,"Bing queries may take me a while ...\n", sep=" ")
 Frequ <- matrix(c(rep(0,rlgth*clgth)), nrow = rlgth)
 colnames(Frequ) <- colkw
 rownames(Frequ) <- rowkw
 urlname1 <- 'https://www.bing.com/search?q='
 # qu <- "%22"
 # qu <-""
 i <- 1
 while(i < rlgth + 1){
 j <- 1    #if (sym) {j <- i} else {j <- 1}  # decomment for symmetrization
  while(j < clgth + 1){
    closeAllConnections()
    tmprkw <- gsub("[[:space:]]", "+", rowkw[i])
    tmpckw <- gsub("[[:space:]]", "+", colkw[j])
    #urlname <- paste(urlname1,rowkw[i],"+",colkw[j],context,sep="")
    #urlname <- paste(urlname1,qu,tmprkw,qu,"+",qu,tmpckw,qu,"+",context,sep="")
    url <- paste(urlname1,tmprkw,"+",tmpckw,"+",context,sep="")
    #if (sym && i == j) {url <- paste(urlname1,rowkw[i],context,sep="")}
    #if (sym && i == j) {url <- paste(urlname1,qu,tmprkw,qu,"+",context,sep="")}
    if (sym && i == j) {url <- paste(urlname1,tmprkw,sep="")}
    
    s <- read_html(url)
    p <- s %>% html_elements("body")
    p <- as(p, "character")
    writeLines(p, 'tmp')
    q <- scan('tmp','character')
    loc <- which(q == 'results</span></div>')
    if (identical(loc, integer(0))) { loc <- which(q == 'Ergebnisse</span></div>') }
    nr <- paste(str_trim(q[loc-3]), str_trim(q[loc-2]), str_trim(q[loc-1]), sep="")
    if (str_length(q[loc-4]) <= 3) { nr <- paste(str_trim(q[loc-4]), nr, sep="") }

    nr <- extract.nr(nr)
    # if (length(loc) == 0) {nr <- 0}
    # print(nr)
    Frequ[i,j] <- Frequ[i,j] + nr
    # if (sym) {Frequ[j,i] <- Frequ[i,j]}   # decomment for symmetrization
    unlink('tmp')
    unlink(url)
    j <- j + 1
  }
 i <- i + 1
 }
 closeAllConnections()
 if (outfn != ""){
  Frequ[is.na(Frequ)] <- 0				# replace NAs by 0
  write(t(Frequ), outfn, ncolumns = clgth, sep=" ")
  cat("\nMatrix written to file ",outfn,":\n",sep="")
 } 
 print(Frequ)
 return(Frequ)
} 

# rownormalize matrix with
row.normalize <- function(rawmat, rowkwfreq, rowfn="", colfn="", outfn=""){
 nrow <- nrow(rawmat)
 ncol <- ncol(rawmat)
 frequ <- matrix(rowkwfreq, nrow, ncol=1)
 tmp <- matrix(rep(frequ, ncol), nrow)
 mat <- rawmat / tmp
 mat[which(mat==Inf)] <- 0
 if (rowfn != ""){
  rownames(mat) <- scan(rowfn, what = 'character')
 }
 if (colfn != ""){ 
  colnames(mat) <- scan(colfn, what = 'character')
 } 
 if (outfn != ""){
  write(t(mat), outfn, ncolumns=ncol, sep=" ")
  cat("\nRow-normalized Matrix written to file ", outfn, ":\n", sep="")
 }
 print(mat)
}  

# capture frequencies of context
### needed for Bing
gsearch.sca <- function(context){
  urlname1 <- 'https://www.bing.com/search?q='
  closeAllConnections()
  url <- paste(urlname1, context, sep="")

  s <- read_html(url)
  p <- s %>% html_elements("body")
  p <- as(p, "character")
  writeLines(p, 'tmp')
  q <- scan('tmp','character')
  loc <- which(q == 'results</span></div>')
  if (identical(loc, integer(0))) { loc <- which(q == 'Ergebnisse</span></div>') }
  nr <- paste(str_trim(q[loc-3]), str_trim(q[loc-2]), str_trim(q[loc-1]), sep="")
  if (str_length(q[loc-4]) <= 3) { nr <- paste(str_trim(q[loc-4]), nr, sep="") }

  nr <- extract.nr(nr)
  unlink('tmp')
  unlink(url)
  closeAllConnections()
  print(nr)
  return(nr)
}

# compute cell-normalized proximities from co-occurrence matrix
cell.normalized.prox <- function(rowcol, rowfreq, colfreq, rowfn="", colfn="", outfn=""){
 nrow <- nrow(rowcol)
 ncol <- ncol(rowcol)
 rowfreq <- matrix(rowfreq, nrow, ncol=1, byrow=TRUE)
 colfreq <- matrix(colfreq, nrow=1, ncol, byrow=TRUE)
 multrow <- matrix(rep(rowfreq, ncol), nrow)
 multcol <- matrix(rep(colfreq, nrow), nrow, byrow=TRUE) 
 nco <- rowcol / (multrow + multcol - rowcol)
 nco <- matrix(rep(0,nrow*ncol), nrow, ncol) + (nco * (nco >= 0))
 if (colfn != ""){
  colkw <- scan(colfn, what = 'character')
  colnames(nco) <- colkw
 } 
 if (rowfn != ""){
  rowkw <- scan(rowfn, what = 'character')
  rownames(nco) <- rowkw
 } 
 #print(nco)
 if (outfn != ""){
  write(t(nco), outfn, ncolumns=ncol, sep=" ")
  cat("\nProximity matrix written to file ", outfn, ":\n", sep="")
  print(nco)
 }
 return(nco)
}

# compute the Normalized Google Distance
# according to Cilibrasi & Vitanyi (2007)
NGD <- function(rowcol, rowfreq, colfreq, N, rowfn="", colfn="", outfn=""){
 nrow <- nrow(rowcol)
 ncol <- ncol(rowcol)
 rowfreq <- matrix(rowfreq, nrow, ncol=1, byrow=TRUE)
 colfreq <- matrix(colfreq, nrow=1, ncol, byrow=TRUE)
 multrow <- matrix(rep(rowfreq, ncol), nrow)
 multcol <- matrix(rep(colfreq, nrow), nrow, byrow=TRUE) 
 minf <- multcol * (multcol <= multrow) + multrow * (multrow < multcol) 
 maxf <- multcol * (multcol >= multrow) + multrow * (multrow > multcol)
 NGD <- (log(maxf) - log(rowcol)) / (log(N) - log(minf))
 NGD <- matrix(rep(0,nrow*ncol), nrow, ncol) + (NGD * (NGD >= 0))
 if (colfn != ""){
  colkw <- scan(colfn, what = 'character')
  colnames(NGD) <- colkw
 } 
 if (rowfn != ""){
  rowkw <- scan(rowfn, what = 'character')
  rownames(NGD) <- rowkw
 } 
 NGD[NGD == Inf] <- -1				# replace Inf by max dist
 NGD[NGD == -1] <- max(NGD)
 #print(NGD)
 if (outfn != ""){
  write(t(NGD), outfn, ncolumns=ncol, sep=" ")
  cat("\nNGD Matrix written to file ", outfn, ":\n", sep="")
  print(NGD)
 }
 return(NGD)
}

# dendrogram for (1 - correlations)
dendro.cor <- function(recmat, rowfn){
 c <- cor(t(recmat))
 rownames(c) <- scan(rowfn, what = 'character')
 dendro <- hclust(as.dist(1-c))
 plot(dendro, main="Dendrogram for (1 - cor)")
} 
   
# dendrogram for distances
dendro.dis <- function(recmat, rowfn){
 d <- dist(recmat, diag=TRUE, upper=TRUE)
 citynames <- scan(rowfn, what = 'character')
 dendro <- hclust(d)
 plot(dendro, main="Dendrogram for distances", labels=citynames)
} 

# MDS (classical metric) for distances
mds.mds <- function(recmat, rowfn, k){
 citynames <- scan(rowfn, what = 'character')
 d <- dist(recmat, diag=TRUE, upper=TRUE)
 citymds <- cmdscale(d, k)
 plot(citymds, pch="", main="Classical metric MDS")
 text(citymds[,1], citymds[,2], citynames)
} 

# MDS (Sammon mapping) 
mds.sam <- function(recmat, rowfn, k){
 library(MASS)
 citynames <- scan(rowfn, what = 'character')
 d <- dist(recmat, diag=TRUE, upper=TRUE)
 citysamm <- sammon(d, y = cmdscale(d, k), k = 2, niter = 100, trace = TRUE, magic = 0.2)
 plot(citysamm$points, pch="", main="MDS (Sammon mapping)")
 text(citysamm$points[,1], citysamm$points[,2], citynames)
}

# write input and run CLIP50.exe
run.clip <- function(recmat, rowfn){
 d <- as.matrix(dist(recmat, diag=TRUE, upper=TRUE))
 write(nrow(d), 'CLIP.in')
 di <- format(as.integer(d * 1000), trim=FALSE, width=4)
 di <- matrix(di,nrow=nrow(d), ncol=ncol(d), byrow=TRUE)
 i <- 0
 while (i < nrow(d)-1){
  i <- i+1
  write(di[i,(i+1):ncol(d)], 'CLIP.in', sep="", ncol=ncol(d)-i, append=TRUE)
 }
 system("CLIP50.exe", wait=TRUE)
 cat("\nCLIP output written on file CLIP.out\n")
 rowkw <- read.table(rowfn)
 colnames(rowkw) <- "point numbers & labels:  "
 print(rowkw)  
} 
